from .defaults.tiers import (
    DEFAULT_TIERS,
    Home,
    WorkPackage,
    Experiment,
    Sample,
    DataSet,
)
from .environment import env
from .core import HomeTierBase, FolderTierBase, NotebookTierBase, Project

try:
    __IPYTHON__  # type: ignore[name-defined]
    from .magics import register

    register()
except NameError:
    pass

__all__ = [
    "Project",
    "DEFAULT_TIERS",
    "env",
    "TierABC",
    "Home",
    "WorkPackage",
    "Experiment",
    "Sample",
    "DataSet",
]



import functools

from typing import (
    Callable,
    Union,
    Any,
    cast,
    Dict,
    List,
    Tuple,
    TypeVar,
    Generic,
    Optional,
    Type,
    overload,
)
from typing_extensions import Self

from .environment import env

T = TypeVar("T")
V = TypeVar("V")


class _SoftProp(Generic[T, V]):
    """
    Like a property, except can be overwritten.
    """

    def __init__(self, func: Callable[[T], V]):
        self.func = func
        self.__wrapped__ = func

    @overload
    def __get__(self, instance: None, owner: Type[T]) -> Self:
        pass

    @overload
    def __get__(self, instance: T, owner: Type[T]) -> V:
        pass

    def __get__(self, instance: Optional[T], owner: Type[T]) -> Union[V, Self]:
        if instance:
            return self.func(instance)
        return self


def soft_prop(wraps: Callable[[Any], V]) -> V:
    """
    Create an over-writable property
    """
    return cast(V, functools.wraps(wraps)(_SoftProp(wraps)))  # type: ignore[arg-type]


class _CachedProp(Generic[T, V]):
    """
    Like a read only property, except it's only evaluated once.

    Cached value is stored in the _CachedProp instance, which is maybe a bad idea, idk.
    """

    def __init__(self, func: Callable[[T], V]):
        self.func = func
        self.cache: Dict[T, V] = env.create_cache()

        self.__wrapped__ = func

    @overload
    def __get__(self, instance: None, owner: Type[T]) -> Self:
        pass

    @overload
    def __get__(self, instance: T, owner: Type[T]) -> V:
        pass

    def __get__(self, instance: Optional[T], owner: Type[T]) -> Union[V, Self]:
        if instance is None:
            return self

        if instance in self.cache:
            return self.cache[instance]

        val = self.func(instance)
        self.cache[instance] = val

        return val

    def __set__(self, instance: Optional[T], value: Any) -> None:
        raise AttributeError("Trying to set a cached property - naughty!")

    @property
    def __isabstractmethod__(self):
        return getattr(self.func, "__isabstractmethod__", False)


def cached_prop(wrapped: Callable[[Any], V]) -> V:
    """
    Decorator for turning functions/ methods into `_CachedProp`s.
    """
    return cast(V, functools.wraps(wrapped)(_CachedProp(wrapped)))  # type: ignore[arg-type]


class _CachedClassProp(Generic[T, V]):
    """
    Like a combination between `classmethod` and `property`.

    The class instance is passed to the wrapped method upon calling.

    Also performs same caching as `_CachedProp`.
    """

    def __init__(self, func: Callable[[Type[T]], V]):
        self.func = func
        self.cache: Dict[Type[T], V] = env.create_cache()

        self.__wrapped__ = func

    def __get__(self, instance: T, owner: Type[T]) -> V:
        if owner in self.cache:
            return self.cache[owner]

        val = self.func(owner)
        self.cache[owner] = val

        return val

    def __set__(self, instance: T, value: Any) -> Any:
        raise AttributeError("Trying to set a cached class property - naughty!")

    @property
    def __isabstractmethod__(self):
        return getattr(self.func, "__isabstractmethod__", False)


def cached_class_prop(wrapped: Callable[[Any], V]) -> V:
    """
    Decorator for turning functions/ methods into `_CachedClassProp`s.

    First argument of wrapped will be `self.__class__` rather than `self`.
    """
    return cast(V, functools.wraps(wrapped)(_CachedClassProp(wrapped)))  # type: ignore[arg-type]



from typing import List, Union, TYPE_CHECKING

from IPython.display import publish_display_data


if TYPE_CHECKING:
    from .core import TierABC, Project


class JLGui:
    """
    Provides UI for interacting with tiers in JupyterLab.

    Parameters
    ----------
    tier: TierABC
        Tier to provide a gui for.

    """

    def __init__(self, tier: "TierABC"):
        self.tier = tier

    def header(self):
        """
        Display header widget.
        """
        publish_display_data({"application/cassini.header+json": {}})

    def meta_editor(self, name: Union[str, List[str]]):
        """
        Display meta editor widget.

        Parameters
        ----------
        name: Union[str, List[str]]
            name or names of the meta attributes to edit.
        """
        if isinstance(name, str):
            name = [name]

        publish_display_data({"application/cassini.metaeditor+json": {"values": name}})


def extend_project(project: "Project"):
    """
    Extend a project to insert JLGui as the `TierBase.gui_class`. Not usually required as the JLGui is the default.

    Parameters
    ----------
    project: Project
        The project to extend.

    Returns
    -------
    project: Project
        The extended project.
    """
    for tier_class in project.hierarchy:
        tier_class.gui_cls = JLGui

    return project



from abc import abstractmethod
from typing import Callable, List

from nbformat import NotebookNode

CellProcessor = Callable[[NotebookNode], None]


class BaseMigrator:
    cell_processors: List[CellProcessor] = []

    def __init_subclass__(cls) -> None:
        cls.cell_processors = []

    @classmethod
    def _cell_processor(cls, f: CellProcessor):
        cls.cell_processors.append(f)
        return f

    def walk_tiers(self):
        yield self.home

        for wp in self.home:
            yield wp

            for exp in wp:
                yield exp

                for smpl in exp:
                    yield smpl

    @abstractmethod
    def migrate(self):
        """
        Perform the migration.
        """


def cell_processor(f):
    return BaseMigrator._cell_processor(f)



from typing import TYPE_CHECKING

from .gui import GUIS
from .components import BaseTierGui


if TYPE_CHECKING:
    from ...core import Project


def extend_project(project: "Project") -> "Project":
    """
    Extend `project` to use the IPython gui.
    """
    for Tier in project.hierarchy:
        gui_cls = GUIS.get(Tier, BaseTierGui)
        Tier.gui_cls = gui_cls

    return project



from typing import Dict, TYPE_CHECKING, Type

from IPython.display import display
from ipywidgets import Button, DOMWidget, HBox, SelectMultiple, Text

from .components import BaseTierGui, SearchWidget, widgetify_html, InputSequence

from cassini.defaults import Home, WorkPackage, Experiment, Sample, DataSet
from cassini.core import TierABC, TierGuiProtocol


class HomeGui(BaseTierGui[Home]):
    def _get_header_components(self) -> Dict[str, DOMWidget]:
        components = dict()
        components["Search"] = lambda: SearchWidget().as_widget()
        child = self.tier.child_cls
        if child:
            child_name = child.pretty_type
            components[f"{child_name}s"] = self._build_children
            components[f"New {child_name}"] = self.new_child

        return components


class ExperimentGui(BaseTierGui[Experiment]):
    def new_dataset(self) -> DOMWidget:
        """
        A handy widget for creating new `DataSets`.
        """
        samples = list(self.tier)
        option_map = {sample.name: sample for sample in samples}

        selection = SelectMultiple(options=option_map.keys(), description="Auto Add")

        def create(form, name, auto_add):
            with form.status:
                self.tier.setup_technique(name)
                if auto_add:
                    for sample in (option_map[name] for name in auto_add):
                        o = sample[name]
                        o.setup_files()
                        display(widgetify_html(o._repr_html_()))

        form = InputSequence(
            create, Text(description="Name:", placeholder="e.g. XRD"), selection
        )

        return form.as_widget()

    def _get_header_components(self) -> Dict[str, DOMWidget]:
        components = super()._get_header_components()
        components["New Data"] = self.new_dataset
        return components


class SampleGui(BaseTierGui[Sample]):
    def _build_children(self) -> DOMWidget:
        buttons = []
        for dataset in self.tier.datasets:
            b = Button(description=dataset.id)

            def make_callback(dataset):
                def open_folder(change):
                    dataset.open_folder()

                return open_folder

            b.on_click(make_callback(dataset))
            buttons.append(b)
        return HBox(tuple(buttons))

    def _get_header_components(self) -> Dict[str, DOMWidget]:
        components = super()._get_header_components()
        return components


GUIS: Dict[Type[TierABC], Type[TierGuiProtocol]] = {
    Home: HomeGui,
    WorkPackage: BaseTierGui[WorkPackage],
    Experiment: ExperimentGui,
    Sample: SampleGui,
    DataSet: BaseTierGui[DataSet],
}



import time
from pathlib import Path
from typing import (
    Any,
    Callable,
    ClassVar,
    Dict,
    Generic,
    KeysView,
    List,
    Optional,
    Mapping,
    TypeVar,
    Union,
    Literal,
)
from typing_extensions import Tuple, cast, Self, Type

from pydantic import (
    BaseModel,
    ConfigDict,
    Field,
    create_model,
    JsonValue,
    ValidationError,
)
from pydantic.fields import FieldInfo


JSONType = TypeVar("JSONType")
AttrType = TypeVar("AttrType")


class MetaCache(BaseModel):
    __pydantic_extra__: Dict[str, JsonValue] = Field(init=False)
    model_config = ConfigDict(
        extra="allow",
        validate_assignment=True,
        revalidate_instances="subclass-instances",
        strict=True,
    )


class MetaValidationError(ValueError):
    def __init__(self, *, validation_error: ValidationError, file: Path):
        message = f"Invalid data in file: {file}, due to the following validation error:\n\n{validation_error}"

        super().__init__(message)
        self.file = file
        self.validation_error = validation_error


class Meta:
    """
    Like a dictionary, except linked to a json file on disk. Caches the value of the json in itself.

    Arguments
    ---------
    file: Path
           File Meta object stores information about.

    Attributes
    ----------
    file: Path
        Path to the meta file.
    model: MetaCache
        Pydantic model representation of this meta object. Used to perform validation, and serial/deserialisation.
    """

    timeout: ClassVar[int] = 1
    my_attrs: ClassVar[List[str]] = ["model", "_cache", "_cache_born", "file"]

    def __init__(self, file: Path, model: Union[Type[MetaCache], None] = None):
        if model is None:
            model = MetaCache

        self._cache_born: float = 0.0
        self.file: Path = file
        self.model: Type[MetaCache] = model

        self._cache: MetaCache = self.model()

    @property
    def age(self) -> float:
        """
        time in secs since last fetch
        """
        return time.time() - self._cache_born

    def fetch(self) -> MetaCache:
        """
        Fetches values from the meta file and updates them into `self._cache`.

        Notes
        -----
        This doesn't *overwrite* `self._cache` with meta contents, but updates it. Meaning new stuff to file won't be
        overwritten, it'll just be loaded.
        """
        if self.file.exists():
            try:
                self._cache = self.model.model_validate_json(
                    self.file.read_text(encoding="utf-8"), strict=False
                )
            except ValidationError as e:
                raise MetaValidationError(validation_error=e, file=self.file)

            self._cache_born = time.time()

        return self._cache

    def refresh(self) -> None:
        """
        Check age of cache, if stale then re-fetch
        """
        if self.age >= self.timeout:
            self.fetch()

    def write(self) -> None:
        """
        Overwrite contents of cache into file
        """
        jsons = self._cache.model_dump_json(
            exclude_defaults=True, exclude={"__pydantic_extra__"}
        )
        with self.file.open("w", encoding="utf-8") as f:
            f.write(jsons)

    def __getitem__(self, item: str) -> Any:
        self.refresh()
        try:
            return getattr(self._cache, item)
        except AttributeError as e:
            raise KeyError(e)

    def __setitem__(self, key: str, value: Any) -> None:
        self.__setattr__(key, value)

    def __getattr__(self, item: str) -> Any:
        self.refresh()
        try:
            return getattr(self._cache, item)
        except KeyError:
            raise AttributeError(item)

    def __setattr__(self, name: str, value: Any) -> None:
        if name in self.my_attrs:
            super().__setattr__(name, value)
        else:
            self.fetch()

            try:
                setattr(self._cache, name, value)
            except ValidationError as e:
                raise MetaValidationError(validation_error=e, file=self.file)

            self.write()

    def __delitem__(self, key: str) -> None:
        self.fetch()
        excluded = self._cache.model_dump(
            exclude={"__pydantic_extra__", key}, exclude_defaults=True
        )
        # it might not be possible for this to happen, because all fields have to have defaults.
        try:
            self._cache = self.model.model_validate(excluded)
        except ValidationError as e:
            raise MetaValidationError(validation_error=e, file=self.file)

        self.write()

    def __repr__(self) -> str:
        self.refresh()
        return f"<Meta {self._cache} ({self.age * 1000:.1f}ms)>"

    def get(self, key: str, default: Any = None) -> Any:
        """
        Like `dict.get`
        """
        try:
            return self.__getattr__(key)
        except AttributeError:
            return default

    def keys(self) -> KeysView[str]:
        """
        like `dict.keys`
        """
        self.refresh()
        return self._cache.model_dump(
            exclude={"__pydantic_extra__"}, exclude_defaults=True
        ).keys()

    @classmethod
    def build_meta_model(thisCls, wrappedCls):
        meta_attrs = {}

        fields = set()

        for cls in wrappedCls.__mro__:
            for name in cls.__dict__:
                if name == "meta_model":
                    continue

                attr = getattr(cls, name)

                if isinstance(attr, MetaAttr):
                    meta_attrs[name] = attr
                    fields.add(attr.as_field())

        cls_name = wrappedCls.__name__

        return create_model(
            f"{cls_name}MetaCache",
            __base__=MetaCache,
            **{name: field for name, field in fields},
        )

    @classmethod
    def create_meta(cls, path: Path, owner: object):
        """
        Create meta object at `path`.

        The appropraite additional fields for each meta attribute are passed on.
        """
        if hasattr(owner, "meta_model") and owner.meta_model:
            model = owner.meta_model
        else:
            model = cls.build_meta_model(owner.__class__)

        return Meta(path, model)


def _null_func(val: Any) -> Any:
    return cast(JsonValue, val)


class MetaAttr(Generic[AttrType, JSONType]):
    """
    Accessor for getting values from a Tier class's meta as an attribute.

    Supports basic serial and de-serialisation.

    Isn't fussy, in that it won't raise an exception if it can't find its key.

    Arguments
    ---------
    post_get: func
        function to apply to data after being loaded from json file
    pre_set: func
        function to apply to data before stored in json file.
    name: str
        key to lookup in meta
    default:
        value to return if key not found in meta (note post_get isn't called on this).

    Examples
    ========

    >>> class MyTier(TierBase):
    >>>
    >>>     list_attr = MetaAttr(post_get=lambda val: val.split(','),
    ...                          pre_set=lambda val: ','.join(val))

    """

    def __init__(
        self,
        json_type: Type[JSONType],
        attr_type: Type[AttrType],
        post_get: Callable[[JSONType], AttrType] = _null_func,
        pre_set: Callable[[AttrType], JSONType] = _null_func,
        name: Union[str, None] = None,
        default: Union[AttrType, None] = None,
        cas_field: Union[None, Literal["core"], Literal["private"]] = None,
    ):
        self.json_type = json_type
        self.attr_type = attr_type

        self.post_get = post_get
        self.pre_set = pre_set

        self.name: str = cast(str, name)
        self.default = default

        self.cas_field = cas_field

    def __set_name__(self, owner: object, name: str) -> None:
        if self.name is None:
            self.name = name

    def __get__(
        self, instance: Union[Any, None], owner: object
    ) -> Union[AttrType, Self]:
        if owner is None or instance is None:
            return self

        return self.post_get(instance.meta.get(self.name, self.default))

    def __set__(self, instance: Any, value: AttrType) -> None:
        setattr(instance.meta, self.name, self.pre_set(value))

    def as_field(self) -> Tuple[str, Tuple[Type[JSONType], FieldInfo]]:
        if self.cas_field:
            return self.name, (
                self.json_type,
                Field(
                    default=self.default,
                    json_schema_extra={"x-cas-field": self.cas_field},
                ),
            )
        else:
            return self.name, (self.json_type, Field(default=self.default))


T = TypeVar("T")



from typing import Sequence, Tuple, Callable

from cassini import (
    Home,
    HomeTierBase,
    NotebookTierBase,
    FolderTierBase,
    DEFAULT_TIERS,
    env,
)
from cassini.core import Project, TierABC

import pytest


ALL_TIERS = [*DEFAULT_TIERS, TierABC, HomeTierBase, FolderTierBase, NotebookTierBase]


@pytest.fixture
def get_Project():
    """
    Get a clean version of the Project class, with caches and other class attributes appropriately reset.
    """
    env._reset()

    Project.__before_setup_files__ = []
    Project.__after_setup_files__ = []
    Project.__before_launch__ = []
    Project.__after_launch__ = []

    return Project


@pytest.fixture
def patch_project(get_Project, tmp_path):
    Project = get_Project

    class Tier(NotebookTierBase):
        pass

    project = Project([Home, Tier], tmp_path)
    return Tier, project


@pytest.fixture
def patched_default_project(
    get_Project, tmp_path
) -> Tuple[Project, Callable[[Sequence[str]], Sequence[TierABC]]]:
    """
    Returns
    -------
    project: Project
        The fresh
    create_tiers: Callable[[Sequence[str]], Sequence[TierABC]]
        function for creating tiers by a list of names. Returns each pre-setup tier in order.
    """
    Project = get_Project

    project = Project(DEFAULT_TIERS, tmp_path)
    project.setup_files()

    def create_tiers(names: Sequence[str]) -> Sequence[TierABC]:
        """
        Provide a list of tiers.

        These tiers will be created in order and all their files will be setup.
        """
        tiers = []

        for name in names:
            tier = project[name]

            parent = tier

            parents = []

            while parent:
                parents.append(parent)
                parent = parent.parent

            for child in reversed(parents):  # work from top down
                try:
                    child.setup_files()
                except FileExistsError:
                    pass

            tiers.append(tier)

        return tiers

    return project, create_tiers



import importlib
from pathlib import Path
import os
import sys
import functools
from typing import (
    MutableMapping,
    Type,
    Union,
    Callable,
    Any,
    List,
    Tuple,
    TypeVar,
    Generic,
)
from jupyterlab.labapp import LabApp, LabServerApp
from typing_extensions import Self, ParamSpec
import datetime

from .environment import env

import jinja2


class PathLibEnv(jinja2.Environment):
    """
    Subclass of `jinja2.Environment` to enable using `pathlib.Path` for template names.
    """

    def get_template(
        self,
        name: Union[Path, str],  # type: ignore[override]
        parent: Union[str, None] = None,
        globals: Union[MutableMapping[str, Any], None] = None,
    ) -> jinja2.Template:
        return super().get_template(
            name.as_posix() if isinstance(name, Path) else name,
            parent=parent,
            globals=globals,
        )


class CassiniLabApp(LabApp):  # type: ignore[misc]
    """
    Subclass of `jupyterlab.labapp.LabApp` that ensures `ContentsManager.allow_hidden = True`
    (needed for jupyter_cassini_server)
    """

    @classmethod
    def initialize_server(
        cls: Type[LabApp], argv: Union[Any, None] = None
    ) -> LabServerApp:
        """
        Patch serverapp to ensure hidden files are allowed, needed for jupyter_cassini_server
        """
        serverapp: LabServerApp = super().initialize_server(argv)
        serverapp.contents_manager.allow_hidden = True
        return serverapp


class FileMaker:
    """
    Utility class for making files, and then rolling back if an exception occurs.
    """

    def __init__(self) -> None:
        self.folders_made: List[Path] = []
        self.files_made: List[Path] = []

    def __enter__(self) -> Self:
        self.files_made = []
        self.folders_made = []
        return self

    def mkdir(self, path: Path, exist_ok: bool = False) -> Union[Path, None]:
        if not path.exists():
            path.mkdir()
            self.files_made.append(path)
            return path
        if exist_ok:
            return None
        raise FileExistsError(path)

    def write_file(
        self, path: Path, contents: str = "", exist_ok: bool = False
    ) -> Union[Path, None]:
        if not path.exists():
            path.write_text(contents, encoding="utf-8")
            self.files_made.append(path)
            return path
        if exist_ok:
            return None
        raise FileExistsError(path)

    def copy_file(self, source: Path, dest: Path) -> Tuple[Path, Path]:
        self.write_file(dest, source.read_text())
        return source, dest

    def __exit__(self, exc_type: Any, exc_val: Any, exc_tb: Any) -> None:
        if exc_type:
            print(f"{exc_type} occured, rolling back")
            for file in self.files_made:
                print("Deleting", file)
                file.unlink()
                print("Done")

            for folder in self.folders_made:
                print("Removing", folder)
                folder.rmdir()
                print("Done")
            raise exc_type(exc_val)


R = TypeVar("R")
P = ParamSpec("P")


class XPlatform(Generic[P, R]):
    """
    Class for easily making cross platform functions/ methods(?). Stops nasty if elses.

    Fallback functional called if no matching function added for current platform.

    Parameters
    ----------
    default: Callable[P, R]

    Returns
    -------
    self: XPlatform[P, R]
        Callable that will always call the appropriate function for the current platform.
    """

    def __init__(self, default: Callable[P, R]) -> None:
        self._func: Union[Callable[P, R], None] = None
        self._default: Callable[P, R] = default

    @property
    def func(self) -> Callable[P, R]:
        """
        Returns the platform appropriate function (or default if not known)
        """
        if self._func:
            return self._func

        return self._default

    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> R:
        return self.func(*args, **kwargs)

    def add(self, platform: str) -> Callable[[Callable[P, R]], "XPlatform[P, R]"]:
        """
        Add a platform dependent function. To be used as a decorator.

        Parameters
        ----------
        platform: str
            platform wrapped function called for.

        Returns
        -------
        self : XPlatform
            Callable that will always call the appropriate function for the current platform.
        """

        def wrapper(func: Callable[P, R]) -> "XPlatform[P, R]":
            if sys.platform == platform:
                self._func = func
                functools.update_wrapper(self, func)
            return self

        return wrapper


@XPlatform
def open_file(filename: Union[str, Path]) -> None:
    """
    *Nix open file implementation.
    """
    import subprocess

    opener = "open" if sys.platform == "darwin" else "xdg-open"
    subprocess.call([opener, filename])


@open_file.add("win32")
def win_open_file(filename: Union[str, Path]) -> None:
    """
    Windows open file implementation.
    """
    os.startfile(filename)  # type: ignore[attr-defined]


def find_project(import_string=None):
    """
    Gets ahold of the Project instance for this Jupyterlab server instance.

    If server was launched via `cassini.Project.launch()`, this will already be set.

    Otherwise, `import_string` or CASSINI_PROJECT environment variable can be set.

    This should be of the form:

        path/to/module:project_obj

    By default, `project_obj` is assumed to be called `project`. This will be imported from `module`, which by default is
    assumed to be `cas_project`.

    Note that for cassini to run with a regular jupyterlab instance, `ContentsManager.allow_hidden = True` must be set, either
     via a config, or passed as a command line argument e.g. `--ContentsManager.allow_hidden=True`
    """
    if env.project:
        return env.project

    if not import_string:
        CASSINI_PROJECT = os.environ["CASSINI_PROJECT"]
    else:
        CASSINI_PROJECT = import_string

    path = Path(CASSINI_PROJECT).absolute()

    module = None
    obj = None

    if ":" in path.name:
        module, obj = path.name.split(":")
        module = module.replace(".py", "")
        directory = path.parent.as_posix()
    elif path.is_file() or path.with_suffix(".py").is_file():
        directory = path.parent.as_posix()
        module = path.stem
        obj = "project"
    elif path.is_dir():
        directory = path.as_posix()
        module = "cas_project"
        obj = "project"
    else:
        raise RuntimeError(f"Cannot parse CASSINI_PROJECT {CASSINI_PROJECT}")

    sys.path.insert(0, directory)

    try:
        env.project = getattr(importlib.import_module(module), obj)
    finally:
        sys.path.remove(directory)

    return env.project



import sys
import os
from typing import Set
from pathlib import Path

from semantic_version import Version, validate, SimpleSpec


def latest_version(cas_lib_dir):
    return max(
        [
            Version(entry.name)
            for entry in os.scandir(cas_lib_dir)
            if validate(entry.name)
        ]
    )


class PatchImporter:
    """
    Context manager that 'patches' in the latest compatible subdirectory of `cas_lib_dir` according to SemVer
    into `sys.path`, allowing modules within this directory to be imported.

    Any patches directories are then removed from `sys.path` upon closure of the context manager.

    Parameters
    ----------
    version: Version
        Version folder to seek.
    cas_lib_dir: Path
        Directory to look into
    """

    _imported: Set[object] = set()

    def __init__(self, version: Version, cas_lib_dir: Path) -> None:
        self._modules_before = None
        spec = SimpleSpec(
            "{major}.{minor}.*".format(major=version.major, minor=version.minor)
        )
        max_compatible = spec.select(
            Version(entry.name)
            for entry in os.scandir(cas_lib_dir)
            if validate(entry.name)
        )

        if max_compatible is None:
            raise ImportError(
                "No compatible version of cas_lib exists to match {}".format(spec)
            )

        print("Using cas_lib version {}".format(max_compatible))

        self.path = str(cas_lib_dir / str(max_compatible))

    def __enter__(self):
        while self._imported:
            del sys.modules[self._imported.pop()]

        sys.path.append(self.path)
        self._modules_before = set(sys.modules)

    def __exit__(self, *args):
        sys.path.remove(self.path)
        self._imported.update(set(sys.modules) - self._modules_before)



import sys
import argparse
from typing import List

from cassini.utils import find_project

from .base import BaseMigrator
from .V0_2toV0_3 import V0_2toV0_3


migrators = {("0.2", "0.3"): V0_2toV0_3}


def main(args: List[str]) -> BaseMigrator:
    parser = argparse.ArgumentParser(
        description=(
            "Migration tool for cassini projects. "
            "It's not that fancy, so sequential updates may be needed."
        )
    )
    parser.add_argument("old", choices=["0.2"], help="which version to migrate from")
    parser.add_argument("new", choices=["0.3"], help="which version to migrate to")
    parser.add_argument(
        "--cassini-project",
        default="project.py:project",
        help="cassini project to migrate, see cassini.utils.find_project for possible forms",
    )

    out = parser.parse_args(args)

    project = find_project(out.cassini_project)

    print("Found project", project)

    old, new = out.old, out.new

    print("Looking for migrator for", old, "to", new)

    try:
        migrator = migrators[(old, new)](project)
    except KeyError:
        raise ValueError(
            "No migator available for this combination, options are", list(migrators)
        )

    print("Found", migrator)

    print("Running migrator")

    migrator.migrate()

    print("Success")

    return migrator


if __name__ == "__main__":
    main(sys.argv[1:])



from __future__ import annotations

from collections import defaultdict
from typing import Any, Dict, List, Union, Optional, Tuple, Generic, TypeVar
from types import MethodType
from typing_extensions import Self
import datetime
from pathlib import Path
import shutil
from io import TextIOWrapper
import warnings

from pydantic import (
    JsonValue,
    BaseModel,
    Field,
    ConfigDict,
    GetCoreSchemaHandler,
)
from pydantic_core import CoreSchema, core_schema

from . import env
from .core import TierABC, NotebookTierBase
from .utils import find_project
from .meta import Meta


class NoseyPath:
    """
    Wrapper for `Path` objects that allows tracking of what paths have been generated from that object.

    This 'nosey' behaviour is used to keep track of which files are accessed through `Path` objects.

    Parameters
    ----------
    path: Path
        path to wrap around.
    """

    @classmethod
    def from_parent(cls, path: Path, parent: Self):
        """
        Create a new `NoseyPath` as a 'child' of `parent`.

        Allows parent to keep track of its child paths.
        """
        obj = cls(path)
        obj._children = parent._children
        parent._children.append(obj)

        return obj

    def __init__(self, path: Path) -> None:
        self._path: Path = path
        self._children: List[Self] = []

    def __getattr__(self, name: str) -> Any:
        val = getattr(self._path, name)

        if isinstance(val, MethodType):

            def wrapper(*args, **kwargs):
                result = val(*args, **kwargs)

                if isinstance(result, Path):
                    result = NoseyPath.from_parent(result, self)

                return result

            return wrapper

        if isinstance(val, Path):
            val = NoseyPath.from_parent(val, self)

        return val

    def __truediv__(self, other: Union[str, Path]) -> Path:
        return self.__getattr__("__truediv__")(other)

    def __eq__(self, other):
        if isinstance(other, NoseyPath):
            other = other._path
        return self._path.__eq__(other)

    def __repr__(self):
        return f"<NoseyPath ({self._path})>"

    def _unchain(self) -> Dict[str, Any]:
        """
        Creates a tree structure that contains all segments of the children of this object.
        """
        segments: Dict[str, Any] = {}

        for child in self._children:
            path = segments
            if child.is_absolute():
                child = child.relative_to(self._path)

            for part in child.parts:
                next_path = path.get(part)

                if next_path is None:
                    next_path = path[part] = {}

                path = next_path

        return segments

    def _recurse_segments(
        self, level: Dict[str, Any], path: List[str], sub_paths: List[List[str]]
    ):
        """
        Recurse through the `level`, keeping track of the `path` and add complete paths to sub_paths.
        """
        for name, branch in level.items():
            path.append(name)

            if not branch:
                sub_paths.append(path.copy())
                path.clear()
            else:
                self._recurse_segments(branch, path, sub_paths)

    def compress(self) -> List[Path]:
        """
        Create a list of complete paths that result from this path object e.g. through `__truediv__` calls.
        """
        if not self._children:
            return [self._path]

        tree = self._unchain()
        path: List[str] = []
        sub_paths: List[List[str]] = []

        self._recurse_segments(tree, path, sub_paths)

        paths = [self._path.joinpath(*sub) for sub in sub_paths]

        return paths


ArgsKwargsType = Tuple[Tuple[Any, ...], Tuple[Tuple[str, Any], ...]]


class SharingTier:
    """
    Wrapper around a `TierABC` object that keeps track of what attributes are accessed
    and what methods are called.

    This class can then create a serialised version of `tier` using `SharedTierData`, which can
    allow a `SharedTier` object to be created that emulates this `SharingTier`.

    This class should not be created directly. Instead, `SharedProject` objects should be used.

    Parameters
    ----------
    name: str
        The name of the tier to wrap around.

    Notes
    -----
    This class is not in a valid state until `SharingTier.load` has been called.
    """

    def __init__(self, name: str):
        self.sharing_project: Union[None, ShareableProject] = None

        self._accessed: Dict[str, Any] = {}
        self._called: Dict[str, Dict[ArgsKwargsType, Any]] = defaultdict(dict)
        self.name = name
        self._paths_used: List[NoseyPath] = []

        self._tier: Union[TierABC, None] = None
        self.meta: Union[Meta, None] = None

    @classmethod
    def with_project(cls, name: str, sharing_project: ShareableProject):
        """
        Create a `SharingTier` object, and load it from `shared_project`.

        Recommended way to create `SharingTier` objects in contexts where the `shared_project` is available.
        """
        tier = cls(name)
        tier.load(sharing_project=sharing_project)

        sharing_project.shared_tiers.append(tier)

        return tier

    def load(self, sharing_project: ShareableProject):
        """
        Sync this `SharingTier` to wrap around the tier with name `self.name` from the `shared_project`.
        """
        self.sharing_project = sharing_project

        self._tier = sharing_project.project[self.name]

        self.meta = getattr(self._tier, "meta", None)

        """
        if isinstance(self._tier, NotebookTierBase) and self.meta:
            # Link this instance's meta attributes to _tier's meta object
            meta_manager: MetaManager = self._tier.__meta_manager__  # type: ignore[attr-defined]
            meta_manager.metas[self] = meta_manager.metas[self._tier]
        """

    meta_model = NotebookTierBase.meta_model

    description = NotebookTierBase.description
    conclusion = NotebookTierBase.conclusion
    started = NotebookTierBase.started

    def handle_attr(self, name: str, val: Any) -> Any:
        """
        Handle attribute access to cache the result appropriately.
        """
        if isinstance(val, (str, int, list, datetime.date, Path)):
            self._accessed[name] = val

        if isinstance(val, Path):
            val = NoseyPath(val)
            self._paths_used.append(val)

        if isinstance(val, TierABC):
            assert self.sharing_project
            val = self._accessed[name] = SharingTier.with_project(
                val.name, self.sharing_project
            )

        return val

    def handle_call(self, method: str, args_kwargs: ArgsKwargsType, val: Any) -> Any:
        """
        Handle call to a method to allow caching of the result.
        """
        if isinstance(val, TierABC):
            assert self.sharing_project
            val = SharingTier.with_project(val.name, self.sharing_project)

        self._called[method][args_kwargs] = val

        if isinstance(val, Path):
            val = NoseyPath(val)
            self._paths_used.append(val)

        return val

    def __getattr__(self, name: str) -> Any:
        if self.sharing_project is None:
            raise RuntimeError(
                "SharingTier attributes can be accessed until `SharingTier.load` is called"
            )

        val = getattr(self._tier, name)

        if isinstance(val, MethodType):
            handle_call = self.handle_call
            meth = val

            def wrapper(*args, **kwargs) -> Any:
                args_kwargs = (args, tuple(kwargs.items()))

                result = meth(*args, **kwargs)

                return handle_call(name, args_kwargs, result)

            val = wrapper
        else:
            val = self.handle_attr(name, val)

        return val

    def __getitem__(self, other) -> Self:
        return self.__getattr__("__getitem__")(other)

    def __truediv__(self, other) -> Path:
        return self.__getattr__("__truediv__")(other)

    def __eq__(self, other):
        if isinstance(other, (SharedTier, SharingTier)):
            return self.name == other.name
        else:
            raise NotImplementedError()

    def __hash__(self):
        return hash(self.name)

    def dump(self, fs: TextIOWrapper) -> SharedTierData:
        """
        Serialise the cached version of the attribute and function calls to wrapped tier.

        Parameters
        ----------
        fs: TextIOWrapper
            Stream to write serialised data to.

        Returns
        -------
        model: SharedTierData
            Model of this instance.
        """

        called = defaultdict(list)

        for meth, calls in self._called.items():
            for args_kwargs, returns in calls.items():
                called[meth.strip("_")].append(
                    {
                        "args": args_kwargs[0],
                        "kwargs": args_kwargs[1],
                        "returns": returns,
                    }
                )

        called_obj = SharedTierCalls(**called)  # type: ignore[arg-type]

        model = SharedTierData(
            **self._accessed, base_path=self.project.project_folder, called=called_obj
        )

        json_str = model.model_dump_json()
        fs.write(json_str)

        return model

    def find_paths(self) -> List[Path]:
        """
        Go through self._accessed, self._called and find the paths and get their full extent.
        """
        paths: List[Path] = []

        for nosey_path in self._paths_used:
            paths.extend(nosey_path.compress())

        return paths


class SharedTierGui:
    """
    Essentially a mock version of TierGui, for use during a shared context.
    """

    def __init__(self, stier: SharedTier):
        self.stier = stier

    def header(self):
        print(
            "Cassini header cannot gui cannot be used in a shared context - check out cassini to set it up for yourself!"
        )

    def meta_editor(self, name=None):
        print(
            "Cassini meta editor cannot be used in a shared context - check out cassini to set it up for yourself!"
        )


class SharedTier:
    """
    A class that emulates a `TierABC` object without needing a full cassini `Project` configured.

    This class is the mirror of `SharingTier` objects, which create the serialised files required
    to load these objects.

    Parameters
    ----------
    name: str
        The name of the tier to emulate.

    Notes
    -----
    This class is not in a valid state until `SharingTier.load` has been called.
    """

    def __init__(self, name: str) -> None:
        self.name = name
        self.shared_project: Union[None, ShareableProject] = None
        self.base_path: Union[Path, None] = None
        self.meta: Union[Meta, None] = None
        self.gui = SharedTierGui(self)

        self._accessed: Dict[str, Any] = {}
        self._called: Dict[str, Dict[ArgsKwargsType, Any]] = {}

    @classmethod
    def with_project(cls, name: str, shared_project: ShareableProject):
        """
        Create a `SharedTier` instance, and load it from `shared_project`.
        """
        tier = cls(name)
        tier.load(shared_project)
        return tier

    def load(self, shared_project: ShareableProject):
        """
        Load the contents of the shared tier into this object from the `shared_project`.
        """
        self.shared_project = shared_project

        folder, meta_file, frozen_file = shared_project.make_paths(self)

        if meta_file.exists():
            self.meta = Meta.create_meta(meta_file, self)

        with open(frozen_file) as fs:
            model = SharedTierData.model_validate_json(fs.read())
            self.base_path = model.base_path

            accessed = model.model_dump(exclude={"called", "base_path"})
            self._accessed = accessed

            raw_called = model.called.model_dump()

            called: Dict[str, Any] = defaultdict(dict)

            for method, calls in raw_called.items():
                if method in ["truediv", "getitem"]:
                    method = f"__{method}__"

                for call in calls:
                    called[method][(call["args"], call["kwargs"])] = call["returns"]

            self._called = called

    meta_model = NotebookTierBase.meta_model

    description = NotebookTierBase.description
    conclusion = NotebookTierBase.conclusion
    started = NotebookTierBase.started

    def adjust_path(self, path: Path) -> Path:
        """
        Correct path to account for new base path.
        """
        assert self.shared_project
        assert self.base_path

        return self.shared_project.requires_path / path.relative_to(self.base_path)

    def process_tier_val(self, val: Any) -> Any:
        if isinstance(val, Path):
            return self.adjust_path(val)

        if isinstance(val, SharedTier):
            assert self.shared_project
            return SharedTier.with_project(val.name, self.shared_project)

        return val

    def __getattr__(self, name: str) -> Any:
        if self.shared_project is None:
            raise RuntimeError(
                "SharedTier attributes can be accessed until `SharedTier.load` is called"
            )

        if name in self._accessed:
            val = self._accessed[name]

            return self.process_tier_val(val)
        else:

            def meth(*args, **kwargs):
                args_kwargs = (args, tuple(kwargs))

                val = self._called[name][args_kwargs]

                return self.process_tier_val(val)

            return meth

    def __getitem__(self, other: Any) -> Self:
        return self.__getattr__("__getitem__")(other)

    def __truediv__(self, other: Any) -> Path:
        return self.__getattr__("__truediv__")(other)

    def __eq__(self, other):
        if isinstance(other, (SharedTier, SharingTier)):
            return self.name == other.name
        else:
            raise NotImplementedError()

    def __hash__(self) -> int:
        return hash(self.name)


class ShareableTierType:
    """
    Pydantic style type for allowing serialisation of Shared and Sharing Tiers.

    See:

    https://docs.pydantic.dev/latest/concepts/types/#handling-third-party-types
    """

    @classmethod
    def __get_pydantic_core_schema__(
        cls, source_type: Any, handler: GetCoreSchemaHandler
    ) -> CoreSchema:
        def validate_from_str(name: str) -> SharedTier:
            return SharedTier(name)  # always load as Shared

        from_str_schema = core_schema.chain_schema(
            [
                core_schema.str_schema(),  # first 'validate' as string
                core_schema.no_info_plain_validator_function(validate_from_str),
            ]
        )

        shared_or_sharing_schema = (
            core_schema.union_schema(  # either Shared or Sharing are valid python-side
                [
                    core_schema.is_instance_schema(SharedTier),
                    core_schema.is_instance_schema(SharingTier),
                ]
            )
        )

        return core_schema.json_or_python_schema(
            json_schema=from_str_schema,
            python_schema=shared_or_sharing_schema,
            serialization=core_schema.plain_serializer_function_ser_schema(
                lambda o: o.name,
                when_used="json",  # dumping to Python should maintain type.
            ),
        )


ReturnType = TypeVar("ReturnType")


class _SharedTierCall(BaseModel, Generic[ReturnType]):
    args: Tuple[JsonValue, ...]
    kwargs: Tuple[Tuple[str, JsonValue], ...]
    returns: ReturnType


TrueDivCall = _SharedTierCall[Path]
GetItemCall = _SharedTierCall[ShareableTierType]
GetChildCall = _SharedTierCall[ShareableTierType]
SharedTierCall = _SharedTierCall[JsonValue]


class SharedTierCalls(BaseModel):
    __pydantic_extra__: Dict[str, List[SharedTierCall]] = Field(init=False)
    model_config = ConfigDict(extra="allow", validate_assignment=True, strict=True)

    truediv: List[TrueDivCall] = Field(default=[])
    getitem: List[GetItemCall] = Field(default=[])
    get_child: List[GetChildCall] = Field(default=[])


class SharedTierData(BaseModel):
    """
    Serialised form of a shared tier.

    Attributes
    ----------
    file: Optional[Path]
        Absolute path to the file for the notebook.
    folder: Optional[Path]
        Absolute path for folder for the tier.
    parent: Optional[str]
        name of the tier's parent
    href: Optional[str]
        tier's href url
    id: Optional[str]
        the tier's id
    identifiers: List[str]
        the tier's identifiers
    meta_file: Optional[Path]
        Absolute path to the meta file.
    base_path: Path
        Base path used when generating URLs.
    called: SharedTierCalls
        Serialised version of calls made to this tier prior to sharing.
    """

    model_config = ConfigDict(extra="allow", validate_assignment=True, strict=True)

    file: Optional[Path] = Field(default=None)
    folder: Optional[Path] = Field(default=None)
    parent: Optional[ShareableTierType] = Field(default=None)
    href: Optional[str] = Field(default=None)
    id: Optional[str] = Field(default=None)
    identifiers: Optional[List[str]] = Field(default=None)
    meta_file: Optional[Path] = Field(default=None)
    base_path: Path

    called: SharedTierCalls


class ShareableProject:
    """
    Shareable version of `Project`. Allows sharing of notebooks that use Cassini with users who don't have
    Cassini set up.

    This class automatically detects if it's being used in a _sharing_ or _shared_ context and returns the appropriate values.
    It does this by trying to use `cassini.utils.find_project` to get ahold of your project instance. If it can't find
    one, it assumes the code context is _shared_. If it does find one, it assumes you are still sharing this project.

    This object can be used as a substitute for a `Project` instance.

    Parameters
    ----------
    import_string: Optional[str]
        If created in a sharing context, this is used to find the `project` object, using the syntax specified
        in `cassini.utils.find_project`.
    location: Optional[Path]
        location to store/ load the shared project data. Defaults to `Path("Shared")`.
    """

    def __new__(cls, *args, **kwargs) -> Self:
        if env.shareable_project:
            warnings.warn(
                "Creating a new instance of SharingProject, when one has already been created. This can create unexpected behaviour."
            )

        obj = super().__new__(cls)
        env.shareable_project = obj
        return obj

    def __init__(
        self, import_string: Union[str, None] = None, location: Union[Path, None] = None
    ) -> None:
        try:
            self.project = find_project(import_string)
        except (RuntimeError, KeyError):
            self.project = None

        self.shared_tiers: List[SharingTier] = []
        self.location = location if location else Path("Shared")

    def env(self, name: str) -> Union[SharedTier, SharingTier]:
        """
        Equivalent to `Project.env`, except will return the appropriate `SharingTier` or `SharedTier`, depending
        on which is appropriate.

        Parameters
        ----------
        name: str
            Name of the tier to get.
        """
        if self.project:
            tier = SharingTier.with_project(name=name, sharing_project=self)
            env.update(tier)
            return tier
        else:
            return SharedTier.with_project(name=name, shared_project=self)

    def __getitem__(self, name: str) -> Union[SharedTier, SharingTier]:
        """
        Equivalent to `Project.env`, except will return the appropriate `SharingTier` or `SharedTier`, depending
        on which is appropriate.

        Parameters
        ----------
        name: str
            Name of the tier to get.
        """
        if self.project:
            return SharingTier.with_project(name=name, sharing_project=self)
        else:
            return SharedTier.with_project(name=name, shared_project=self)

    def make_paths(
        self, tier: Union[SharedTier, SharingTier]
    ) -> Tuple[Path, Path, Path]:
        """
        Build paths required for sharing a given tier object.

        Returns
        -------
        outer: Path
            The folder the tier data is stored in.
        meta_file: Path
            A path to the meta_file of that tier
        frozen_file: Path
            The path to the `frozen.json` file, where the tier's attributes and calls are stored.
        """
        outer = self.location / tier.name

        meta_file = outer / f"{tier.name}.json"
        frozen_file = outer / "frozen.json"

        return outer, meta_file, frozen_file

    @property
    def requires_path(self) -> Path:
        """
        Path for requires folder.
        """
        return self.location / "requires"

    def make_shared(self) -> None:
        """
        Create a shared version of this project.

        Will create a folder a `self.location`. Will then iterate all `tier` objects accessed in this context
        and serialise them.

        Additionally, and files that tiers used access will be copied into the `self.requires_path`.
        """
        if not self.project:
            raise RuntimeError("Trying to share tiers when not in a sharing context.")

        project = self.project

        path = self.location

        print("Creating shared directory:", path)

        path.mkdir(exist_ok=True)

        print("Success")
        print("Making Requires directory")

        self.requires_path.mkdir(exist_ok=True)

        print("Success")

        for stier in self.shared_tiers:
            print(f"Creating shared version of {stier.name}")

            tier_dir, meta_file, frozen_file = self.make_paths(stier)
            tier_dir.mkdir(exist_ok=True)

            if stier.meta:
                print("Copying Meta")
                shutil.copy(stier.meta.file, meta_file)
                print("Success")

            with open(frozen_file, "w") as fs:
                print("Freezing attributes/ calls")
                stier.dump(fs)
                print("Success")

            print("Making a copy of required files")
            for required in stier.find_paths():
                if required.exists():
                    destination = self.requires_path / required.relative_to(
                        project.project_folder
                    )
                    directory = destination if required.is_dir() else destination.parent
                    directory.mkdir(exist_ok=True, parents=True)

                    print("Copying", required)
                    shutil.copy(required, destination)
                    print("Success")



from typing import Union, TypeVar, Type
from typing_extensions import Annotated
from pathlib import Path

from semantic_version import Version

from .import_tools import PatchImporter, latest_version
from ... import Project, NotebookTierBase
from ...meta import MetaAttr


def extend_project(project: Project, cas_lib_dir: Union[str, Path] = "cas_lib"):
    """
    Extend project to add `cas_lib` attribute to Tiers.
    """
    cas_lib_dir = project.project_folder / cas_lib_dir

    def make_cas_lib_folder(project, cas_lib_dir=cas_lib_dir):
        if not cas_lib_dir.exists():
            cas_lib_dir.mkdir()
            (cas_lib_dir / "0.1.0").mkdir()

    project.__before_setup_files__.append(
        make_cas_lib_folder
    )  # ensures will run even if project already setup.

    for Tier in project.hierarchy:
        if issubclass(Tier, NotebookTierBase):
            # patch in the requried attributes to classes with notebooks!
            Tier.cas_lib_version = MetaAttr(  # type: ignore[attr-defined]
                json_type=str,
                attr_type=Version,
                post_get=lambda v: Version(v) if v else v,
                pre_set=str,
                name="cas_lib_version",
                cas_field="private",
            )

            Tier.cas_lib = create_cas_lib(cas_lib_dir)  # type: ignore[attr-defined]
    return project


def create_cas_lib(cas_lib_dir: Path):
    """
    Create `cas_lib` attribute for a given `cas_lib_dir`.
    """

    def cas_lib(self, version=None):
        """
        Allow importing from the appropraite `cas_lib` subdirectory for this tier.

        If the version not provided or `tier.cas_lib_version` not set, is set to the
        latest version found in `cas_lib_dir`, which is then stored in the tier's meta.

        As `cas_lib_version` is persistent after being set the first time, this will pin this version,
        unless either an explicit `version` parameter is provided, or `tier.cas_lib_version` is updated.
        """
        if version is None:
            if self.cas_lib_version:
                version = self.cas_lib_version
            else:
                version = self.cas_lib_version = latest_version(cas_lib_dir)
                print(f"Set {self}.cas_lib_version = {version}")

        if version == "lastest":
            version = latest_version(cas_lib_dir)

        if isinstance(version, str):
            version = Version(version)

        return PatchImporter(version, cas_lib_dir)

    return cas_lib



from __future__ import annotations

from typing import Union, TYPE_CHECKING, TypeVar, Any, List, Dict
from typing_extensions import TypeGuard

if TYPE_CHECKING:
    from .core import TierABC, Project
    from .sharing import ShareableProject


class _Env:
    """
    Essentially a global object that describes the state of the project for this interpreter.

    As each notebook has its own interpreter, each notebook also has its own env that basically stores what the current
    project is and what tier this `ipynb` file corresponds to.

    Attributes
    ----------
    project : Project
        reference to the current project object. Returns `None` if one not set yet.

    Warnings
    --------
    This object is a singleton, so only one instance can exist at a time.

    This object shouldn't be created directly, instead you should call `project.env('...')` to set its value.
    """

    instance: Union[_Env, None] = None

    def __new__(cls, *args: Any, **kwargs: Any) -> _Env:
        if cls.instance:
            raise RuntimeError(
                "Attempted to create new _Env instance, only 1 _instance permitted per interpreter"
            )
        instance = object.__new__(cls)
        cls.instance = instance
        return instance

    def __init__(self) -> None:
        self.project: Union[Project, None] = None
        self._o: Union[TierABC, None] = None
        self.shareable_project: Union[ShareableProject, None] = None
        self._caches: List[Dict[Any, Any]] = []

    @staticmethod
    def is_sharing(instance: _Env) -> TypeGuard["_SharingInstance"]:
        return bool(instance.shareable_project and instance.project)

    @staticmethod
    def is_shared(instance: _Env) -> TypeGuard["_SharedInstance"]:
        return bool(instance.shareable_project and not instance.project)

    @property
    def o(self) -> Union[TierABC, None]:
        """
        Reference to current Tier object.
        """
        return self._o

    def update(self, obj: TierABC) -> None:
        self._o = obj

    def create_cache(self):
        """
        Method for creating various caches throughout cassini.

        the env instances keeps track of these, to allow them to be cleared
        cleanly during testing.

        This is an internal feature.
        """
        cache = dict()
        self._caches.append(cache)
        return cache

    def _reset(self):
        """
        Reset env instance to initial state. Used for testing.

        `project`, `shareable_project` and `caches` are cleared.
        """
        self.shareable_project = None
        self.project = None

        for cache in self._caches:
            cache.clear()


class _SharingInstance(_Env):
    shareable_project: ShareableProject
    project: Project


class _SharedInstance(_Env):
    shareable_project: ShareableProject
    project: None


env = _Env()



from pathlib import Path
import os

from typing import Iterator, List, Any, Sequence, cast

from ..core import TierABC, FolderTierBase, NotebookTierBase, HomeTierBase
from ..accessors import cached_prop
from ..utils import FileMaker


def ignore_dir(name: str) -> bool:
    if name.startswith("."):
        return True
    if name.startswith("_"):
        return True
    return False


class Home(HomeTierBase):
    """
    Home `Tier`.

    This, or a subclass of this should generally be the first entry in your hierarchy, essentially represents the top
    level folder in your hierarchy.

    Creates the `Home.ipynb` notebook that allows easy navigation of your project.
    """

    pass


class WorkPackage(NotebookTierBase):
    """
    WorkPackage Tier.

    Intended to contain all the work towards a particular goal. i.e. prove that we can do this.

    Next level down are `Experiment`s.
    """

    name_part_template = "WP{}"
    short_type = "wp"

    @property
    def exps(self) -> Sequence[TierABC]:
        """
        Gets a list of all this `WorkPackage`s experiments.
        """
        return list(self)


class Experiment(NotebookTierBase):
    """
    Experiment `Tier`.

    Just below `WorkPackage`, experiments are intended to be collections of samples and datasets that work towards the
    goal of the parent `WorkPackage`.

    Each `Experiment` has a number of samples.
    """

    name_part_template = ".{}"
    short_type = "exp"

    @property
    def techniques(self) -> Sequence[str]:
        """
        Convenience property for looking up all the techniques that have been performed on samples in this experiment.

        Notes
        -----
        This just checks for the existence of DataSet folders, and not if they have anything in them!
        """
        techs = []
        for entry in os.scandir(self.folder):
            if entry.is_dir() and not ignore_dir(entry.name):
                techs.append(entry.name)
        return techs

    def setup_technique(self, name: str) -> None:
        """
        Convenience method for adding a new technique to this experiment.

        Essentially just creates a new folder for it in the appropriate location.

        This folder can then be filled with `DataSet`s
        """
        print("Making Data Folder")
        folder = self.folder / name

        if folder.exists():
            raise FileExistsError(f"{folder} exists already")

        with FileMaker() as maker:
            maker.mkdir(folder)

        print("Done")

    @property
    def smpls(self) -> Sequence[TierABC]:
        """
        Get a list of this `Experiment`s samples.
        """
        return list(self)


class Sample(NotebookTierBase):
    """
    Sample `Tier`.

    A `Sample` is intended to represent some object that you collect data on.

    As such, each sample has its own `DataSet`s.

    Notes
    -----
    A `Sample` id can't start with a number and can't contain `'-'` (dashes), as these confuse the name parser.
    """

    name_part_template = "{}"
    id_regex = r"([^0-9^-][^-]*)"

    @cached_prop
    def folder(self) -> Path:
        assert self.parent
        return self.parent.folder

    @property
    def datasets(self) -> Sequence[TierABC]:
        """
        Convenient way of getting a list of `DataSet`s this sample has.
        """
        assert self.parent
        assert self.child_cls
        assert isinstance(self.parent, Experiment)

        techs = []
        for technique in self.parent.techniques:
            dataset = self.child_cls(*self.identifiers, technique, project=self.project)
            if dataset.exists():
                techs.append(dataset)
        return techs

    def __iter__(self) -> Iterator[TierABC]:
        return iter(self.datasets)


class DataSet(FolderTierBase):
    """
    `DataSet` Tier.

    The final tier, intended to represent a folder containing a collection of files relating to a particular `Sample`.
    """

    short_type = "dset"
    name_part_template = "-{}"

    id_regex = r"(.+)"

    @cached_prop
    def folder(self) -> Path:
        assert self.parent

        return self.parent / self.id / self.parent.id

    def exists(self) -> bool:
        return self.folder.exists()

    def __truediv__(self, other: Any) -> Path:
        return cast(Path, self.folder / other)

    def __iter__(self) -> Iterator["os.DirEntry[Any]"]:
        """
        Call `os.scandir` on `self.folder`.
        """
        yield from os.scandir(self.folder)

    def __fspath__(self) -> str:
        return self.folder.__fspath__()


DEFAULT_TIERS = [Home, WorkPackage, Experiment, Sample, DataSet]



import json
import datetime
import shutil

from .base import BaseMigrator


class V0_2toV0_3(BaseMigrator):
    def __init__(self, project) -> None:
        self.project = project
        self.home = project.home

    def migrate(self):
        from cassini import NotebookTierBase

        for tier in self.walk_tiers():
            if not isinstance(tier, NotebookTierBase):
                continue
            else:
                backup_path = shutil.copy(
                    tier.meta_file, tier.meta_file.with_suffix(".backup")
                )
                try:
                    self.migrate_meta(tier)
                except Exception:
                    raise RuntimeError(
                        "Error occured, please restore from backup", backup_path
                    )
                else:
                    print("Successfully migrated, removing backup")
                    backup_path.unlink()

    def migrate_meta(self, tier):
        with open(tier.meta_file, "r") as fs:
            meta = json.load(fs)

        started = meta.get("started")

        if started:
            started_dt = datetime.datetime.strptime(
                started, "%d/%m/%Y"
            ).astimezone()  # assume local timezone.
            meta["started"] = str(started_dt)

        with open(tier.meta_file, "w") as fs:
            json.dump(meta, fs)



from warnings import warn

from IPython.core.magic import register_cell_magic
from IPython.core.interactiveshell import InteractiveShell
from IPython.display import Markdown, publish_display_data  # type: ignore[attr-defined]

from .environment import env
from .core import NotebookTierBase


def hlt(line: str, cell: str):
    if env.is_shared(env):
        warn(
            "This notebook is in a shared context and therefore highlights magics won't work"
        )
        return cell

    if not isinstance(env.o, NotebookTierBase):
        raise ValueError(
            "Highlights can only be added to tiers that subclass NotebookTierBase"
        )

    if not line:
        raise ValueError("Please provide a title for the highlight")

    assert env.o

    outputs = []

    def capture_display(msg):
        if msg["content"]["data"] in outputs:  # display only once
            return None
        outputs.append(msg["content"])
        return msg

    cell = cell.strip()

    annotation = None

    if cell.endswith('"""'):
        *rest, annotation = cell.split('"""')[:-1]
        cell = '"""'.join(rest)

    shell = InteractiveShell.instance()
    # Capture any output that is displayed before output (like matplotlib plots)
    shell.display_pub.register_hook(capture_display)

    header = f"## {line}"

    try:
        publish_display_data({"text/markdown": header})

        result = shell.run_cell(cell).result

        if shell.display_formatter:
            outputs.append(
                dict(zip(("data", "metadata"), shell.display_formatter.format(result)))
            )

        if annotation:
            publish_display_data({"text/markdown": annotation})
    finally:
        shell.display_pub.unregister_hook(capture_display)

    all_out = outputs

    env.o.add_highlight(line, all_out)

    return None


def conc(line, cell):
    if not isinstance(env.o, NotebookTierBase):
        raise ValueError(
            "Highlights can only be added to tiers that subclass NotebookTierBase"
        )

    if env.o.conclusion and line != "force":
        raise RuntimeError(
            f"Conclusion for {env.o} already set, use %%conc force to force update"
        )
    env.o.conclusion = cell
    return Markdown(cell)


def register():
    register_cell_magic(hlt)
    register_cell_magic(conc)



from __future__ import annotations

import datetime
import html
import json
import os
from pathlib import Path
from abc import ABC, abstractmethod
import re

from typing import (
    Any,
    List,
    Sequence,
    Type,
    Tuple,
    Iterator,
    Union,
    Dict,
    ClassVar,
    Optional,
    Callable,
    cast,
    Protocol,
)
from warnings import warn
from jupyterlab.labapp import LabApp  # type: ignore[import-untyped]
from typing_extensions import Self

import jinja2
from pydantic import JsonValue, AwareDatetime

from .meta import Meta, MetaAttr
from .accessors import cached_prop, cached_class_prop, soft_prop
from .utils import (
    FileMaker,
    open_file,
    CassiniLabApp,
    PathLibEnv,
)
from .environment import env
from .config import config
from .jlgui import JLGui


class TierGuiProtocol(Protocol):
    """
    Protocol for providing a gui for tiers.

    Must provide a header method and take the tier to provide a gui for as the first argument.
    """

    def __init__(self, tier: TierABC):
        pass

    def header(self):
        """
        The header is the UI that is goes at the top of a notebook.
        """
        pass


MetaDict = Dict[str, JsonValue]


HighlightType = List[Dict[str, Dict[str, Any]]]
HighlightsType = Dict[str, HighlightType]


class TierABC(ABC):
    """
    Abstract Base class for creating Tiers objects. Tiers should correspond to a folder on your disk.

    Instances of this class or subclasses should not be created directly. Instead a Project instance
    should create them.

    Parameters
    ----------
    *identifiers : str
        sequence of strings that identify this tier. With the final identifier being unique.

    Attributes
    ----------
    id_regex: str
        Class attribute, regex that defines a group that matches the id of a `Tier` object from a name... except the
        name isn't the full name, but with parent names stripped, see examples basically!
    hierarchy : list
        Class attribute, hierarchy of `Tier`s.
    id_regex : str
        (class attribute) regex used to restrict form of `Tier` object ids. Should contain 1 group that captures the id.
    gui_cls : Any
        (class attribute) The class called upon initialisation to use as gui for this object. Constructor should take `self`
         as first argument.
    """

    _cache: ClassVar[Dict[Tuple[str, ...], TierABC]] = env.create_cache()

    def __init_subclass__(cls, *args: Any, **kwargs: Any) -> None:
        super().__init_subclass__(*args, **kwargs)
        cls._cache = env.create_cache()  # ensures each TierBase class has its own cache

    id_regex: ClassVar[str] = r"(\d+)"

    gui_cls: Type[TierGuiProtocol]

    @cached_class_prop
    def _pretty_type(cls) -> str:
        """
        Name used to display this Tier. Defaults to `cls.__name__`.
        """
        return cls.__name__  # type: ignore[attr-defined]

    pretty_type: str = _pretty_type  # to please type checker.

    @cached_class_prop
    def _short_type(cls) -> str:
        """
        Name used to programmatically refer to instances of this `Tier`.

        Default, take pretty type, make lowercase and remove vowels.
        """
        return cls.pretty_type.lower().translate(str.maketrans(dict.fromkeys("aeiou")))

    short_type: str = _short_type

    @cached_class_prop
    def _name_part_template(cls) -> str:
        """
        Python template that's filled in with `self.id` to create segment of the `Tier` object's name.
        """
        return cls.pretty_type + "{}"

    name_part_template = _name_part_template

    @cached_class_prop
    def _name_part_regex(cls) -> str:
        """
        Regex where first group matches `id` part of string. Default is fill in `cls.name_part_template` with
        `cls.id_regex`.
        """
        return cls.name_part_template.format(cls.id_regex)

    name_part_regex = _name_part_regex

    @classmethod
    @abstractmethod
    def iter_siblings(cls, parent: TierABC) -> Iterator[TierABC]:
        """
        Provide an iterator for the siblings of this tier for a given parent i.e.
        iterate over parent's children.
        """
        pass

    def __new__(cls, *args: str, **kwargs: Dict[str, Any]) -> TierABC:
        obj = cls._cache.get(args)
        if obj:
            return obj
        obj = object.__new__(cls)
        cls._cache[args] = obj
        return obj

    _identifiers: Tuple[str, ...]
    gui: TierGuiProtocol

    def __init__(self: Self, *args: str, project: Project):
        self.project = project

        self._identifiers = tuple(filter(None, args))
        self.gui = self.gui_cls(self)

        rank = self.project.rank_map[self.__class__]

        if len(self._identifiers) != rank:
            raise ValueError(
                f"Invalid number of identifiers in {self._identifiers}, expecting {rank}."
            )

        if self.parse_name(self.name) != self.identifiers:
            raise ValueError(
                f"Invalid identifiers - {self._identifiers}, resulting name ('{self.name}') not in a parsable form "
            )

    @cached_prop
    def _parent_cls(self) -> Union[Type[TierABC], None]:
        """
        `Tier` above this `Tier`, `None` if doesn't have one
        """
        return self.project.get_parent_cls(self.__class__)

    parent_cls = _parent_cls

    @cached_prop
    def _child_cls(self) -> Union[Type[TierABC], None]:
        """
        `Tier` below this `Tier`, `None` if doesn't have one
        """
        return self.project.get_child_cls(self.__class__)

    child_cls = _child_cls

    def parse_name(self, name: str) -> Tuple[str, ...]:
        """
        Ask `env.project` to parse name.
        """
        return self.project.parse_name(name)

    @abstractmethod
    def setup_files(
        self, template: Union[Path, None] = None, meta: Optional[MetaDict] = None
    ) -> None:
        """
        Create all the files needed for a valid `Tier` object to exist.

        This includes its `.ipynb` file, its parent folder, its own folder and its `Meta` file.

        Will render `.ipynb` file as Jinja template engine, passing to the template `self` with names given by
        `self.short_type` and `tier`.

        Parameters
        ----------
        template : Path
            path to template file to render to create `.ipynb` file.
        meta : MetaDict
            Initial meta values to create the tier with.
        """
        pass

    @cached_prop
    def identifiers(self) -> Tuple[str, ...]:
        """
        Read only copy of identifiers that make up this `Tier` object.
        """
        return self._identifiers

    @cached_prop
    def name(self) -> str:
        """
        Full name of `Tier` object. Made by concatenating each parent's `self.name_part_template` filled with each parent's
        `self.id`.

        Examples
        --------

            >>> wp = WorkPackage('2')
            >>> exp = Experiment('2', '3')
            >>> smpl = Sample('2', '3', 'c')
            >>> wp.name_part_template.format(wp.id)
            WP2
            >>> exp.name_part_template.format(exp.id)
            .3
            >>> smpl.name_part_template.format(smpl.id)
            c
            >>> smpl.name  # all 3 joined together
            WP2.3c
        """
        return "".join(
            cls.name_part_template.format(id)
            for cls, id in zip(self.project.hierarchy[1:], self.identifiers)
        )

    @property
    @abstractmethod
    def folder(self) -> Path:
        """
        The folder this tier corresponds to.
        """
        pass

    def open_folder(self) -> None:
        """
        Open `self.folder` in explorer

        Notes
        -----
        Only works on Windows machines.

        Window is opened via the Jupyter server, not the browser, so if you are not accessing jupyter on localhost then
        the window won't open for you!
        """
        open_file(self.folder)

    @cached_prop
    def id(self) -> str:
        """
        Shortcut for getting final identifier.

        Examples
        --------

            >>> from my_project import project
            >>> smpl = project.env('WP2.3c')
            >>> smpl.identifiers
            ['2', '3', 'c']
            >>> smpl.id
            'c'
        """
        return self._identifiers[-1]

    @cached_prop
    def parent(self) -> Union[TierABC, None]:
        """
        Parent of this `Tier` _instance, `None` if has no parent :'(
        """
        if self.parent_cls:
            return self.parent_cls(*self._identifiers[:-1], project=self.project)
        return None

    @cached_prop
    @abstractmethod
    def href(self) -> Union[str, None]:
        """
        href usable in notebook HTML giving link to `self.file`.

        Assumes that `os.getcwd()` reflects the directory of the currently opened `.ipynb` (usually true, unless you're
        changing working dir).

        Does do escaping on the HTML, but is maybe pointless!

        Returns
        -------
        href : str
            href usable in notebook HTML.
        """
        pass

    @abstractmethod
    def exists(self) -> bool:
        """
        returns True if this `Tier` object has already been setup (e.g. by `self.setup_files`)
        """
        pass

    def get_child(self, id: str) -> TierABC:
        """
        Get a child according to the given `id`.

        Parameters
        ----------
        id : str
            id to add `self.identifiers` to form new `Tier` object of tier below.

        Returns
        -------
        child : Type[TierBase]
            child `Tier` object.
        """
        assert self.child_cls
        return self.child_cls(*self._identifiers, id, project=self.project)

    def __truediv__(self, other: Any) -> Path:
        return cast(Path, self.folder / other)

    def __getitem__(self, item: str) -> TierABC:
        """
        Equivalent to `self.get_child(item)`.
        """
        return self.get_child(item)

    def __iter__(self) -> Iterator[Any]:
        """
        Iterates over all children (in no particular order). Children are found by looking through the child meta
        folder.

        Empty iterator if no children.
        """
        if not self.child_cls:
            raise NotImplementedError()

        yield from self.child_cls.iter_siblings(self)

    def __repr__(self) -> str:
        return f'<{self.__class__.__name__} "{self.name}">'

    def _repr_html_(self) -> str:
        block = f'h{len(self.identifiers) + 1} style="display: inline;"'
        return (
            f'<a href="{self.href}"'
            f' target="_blank"><{block}>{html.escape(self.name)}</{block}</a>'
        )

    @abstractmethod
    def remove_files(self) -> None:
        """
        Deletes files associated with a `Tier`
        """
        pass


class FolderTierBase(TierABC):
    """
    Base class for a tier which has a folder, but not notebook/ meta.
    """

    gui_cls = JLGui

    @classmethod
    def iter_siblings(cls, parent: TierABC) -> Iterator[TierABC]:
        # TODO: shouldn't project also handle this?
        if not parent.folder.exists():
            return

        for folder in os.scandir(parent.folder):
            if not folder.is_dir():
                continue
            yield cls(*parent.parse_name(folder.name), project=parent.project)

    @cached_prop
    def folder(self) -> Path:
        """
        Path to folder where the contents of this `Tier` lives.

        Defaults to `self.parent.folder / self.name`.
        """
        if self.parent:
            return self.parent.folder / self.name
        else:  # this is bad
            return Path(self.name)

    def setup_files(self, template: Union[Path, None] = None, meta=None) -> None:
        print(f"Creating Folder for {self} at {self.folder}")

        with FileMaker() as maker:
            maker.mkdir(self.folder.parent, exist_ok=True)
            maker.mkdir(self.folder)

        print("Success")

    def remove_files(self) -> None:
        pass

    def exists(self) -> bool:
        """
        returns True if this `Tier` object has already been setup (e.g. by `self.setup_files`)
        """
        return self.folder.exists()

    @cached_prop
    def href(self) -> Union[str, None]:
        """
        href usable in notebook HTML giving link to `self.file`.

        Assumes that `os.getcwd()` reflects the directory of the currently opened `.ipynb` (usually true, unless you're
        changing working dir).

        Does do escaping on the HTML, but is maybe pointless!

        Returns
        -------
        href : str
            href usable in notebook HTML.
        """
        return html.escape(Path(os.path.relpath(self.folder, os.getcwd())).as_posix())


class NotebookTierBase(FolderTierBase):
    """
    Base class for tiers which have a notebook and meta associated with them.
    """

    meta: Meta

    @cached_class_prop
    def meta_model(cls):
        return Meta.build_meta_model(cls)

    @cached_class_prop
    def _default_template(cls) -> Path:
        """
        Template used to render a tier file by default.
        """
        return Path(cls.pretty_type) / f"{cls.pretty_type}.tmplt.ipynb"

    default_template = _default_template

    @cached_class_prop
    def _meta_folder_name(cls) -> str:
        """
        Form of meta folder name. (Just fills in `config.META_DIR_TEMPLATE` with `cls.short_type`).
        """
        return config.META_DIR_TEMPLATE.format(cls.short_type)

    meta_folder_name = _meta_folder_name

    @classmethod
    def iter_siblings(cls, parent):
        meta_folder = parent.folder / config.META_DIR_TEMPLATE.format(cls.short_type)

        if not meta_folder.exists():
            return

        for meta_file in os.scandir(meta_folder):
            if not meta_file.is_file() or not meta_file.name.endswith(".json"):
                continue
            yield cls(
                *parent.parse_name(meta_file.name[:-5]), project=parent.project
            )  # I don't like this.

    def __init__(self, *identifiers: str, project: Project):
        super().__init__(*identifiers, project=project)
        self.meta: Meta = Meta.create_meta(self.meta_file, owner=self)

    def setup_files(
        self, template: Union[Path, None] = None, meta: Optional[MetaDict] = None
    ) -> None:
        """
        Create all the files needed for a valid `Tier` object to exist.

        This includes its `.ipynb` file, its parent folder, its own folder and its `Meta` file.

        Will render `.ipynb` file as Jinja template engine, passing to the template `self` with names given by
        `self.short_type` and `tier`.

        Parameters
        ----------
        template : Path
            path to template file to render to create `.ipynb` file.
        meta : MetaDict
            Initial meta values to create the tier with.
        """
        if template is None:
            template = self.default_template

        if meta is None:
            meta = {}

        if self.exists():
            raise FileExistsError(f"Meta for {self.name} exists already")

        if self.file and self.file.exists():
            raise FileExistsError(f"Notebook for {self.name} exists already")

        print(f"Creating files for {self.name}")

        print(f"Meta ({self.meta_file})")

        with FileMaker() as maker:
            maker.mkdir(self.meta.file.parent, exist_ok=True)
            maker.write_file(self.meta.file, json.dumps(meta))

            print("Writing Meta Data")

            self.started = datetime.datetime.now(datetime.timezone.utc)

            print("Success")

            print(f"Creating Tier File ({self.file}) using template ({template})")
            maker.write_file(self.file, self.render_template(template))
            print("Success")

            print(f"Creating Tier Folder ({self.folder})")

            self.folder.mkdir(exist_ok=True)

            print("Success")

        print("All Done")

    def exists(self) -> bool:
        """
        returns True if this `Tier` object has already been setup (e.g. by `self.setup_files`)
        """
        return bool(self.file and self.folder.exists() and self.meta_file.exists())

    description = MetaAttr(str, str, cas_field="core")
    conclusion = MetaAttr(str, str, cas_field="core")
    started = MetaAttr(AwareDatetime, datetime.datetime, cas_field="core")

    @cached_prop
    def meta_file(self) -> Path:
        """
        Path to where meta file for this `Tier` object should be.

        Returns
        -------
        meta_file : Path
            Defaults to `self.parent.folder / self.meta_folder_name / (self.name + '.json')`
        """
        assert self.parent
        return self.parent.folder / self.meta_folder_name / (self.name + ".json")

    @cached_prop
    def highlights_file(self) -> Union[Path, None]:
        """
        Path to where highlights file for this `Tier` object should be.

        Returns
        -------
        highlights_file : Path
            Defaults to `self.parent.folder / self._meta_folder_name / (self.name + '.hlts')`
        """
        assert self.parent
        return self.parent.folder / self._meta_folder_name / (self.name + ".hlts")

    @cached_prop
    def file(self) -> Path:
        """
        Path to where `.ipynb` file for this `Tier` instance will be.

        Returns
        -------
        file : Path
            Defaults to self.parent.folder / (self.name + '.ipynb').
        """
        assert self.parent
        return self.parent.folder / (self.name + ".ipynb")

    @classmethod
    def get_templates(cls, project: Project) -> List[Path]:
        """
        Get all the templates for this `Tier`.
        """
        return [
            Path(cls.pretty_type) / entry.name
            for entry in os.scandir(project.template_folder / cls.pretty_type)
            if entry.is_file()
        ]

    def render_template(self, template_path: Path) -> str:
        """
        Render template file passing `self` as `self.short_type` and as `tier`.

        Parameters
        ----------
        template : Path
            path to template file. Must be relative to `project.template_folder`

        Returns
        -------
        rendered_text : str
            template rendered with `self`.
        """
        template = self.project.template_env.get_template(template_path)
        return template.render(**{self.short_type: self, "tier": self})

    def get_highlights(self) -> Union[HighlightsType, None]:
        """
        Get dictionary of highlights for this `Tier` _instance.

        This dictionary is in a form that can be rendered in the notebook using `IPython.display.publish_display_data`
        see Examples.

        This is implemented for you with `self.display_highlights()`.

        Returns
        -------
        highlights : dict
            Get dictionary of highlights for this `Tier` _instance. If the highlights file doesn't exist, just returns an
            empty dict.

        Examples
        --------

            >>> wp = project['WP1']
            >>> for title, outputs in wp.get_highlights():
            ...     print("Displaying Highlight:", title)
            ...     for output in outputs:
            ...         IPython.display.publish_display_data(**output)
        """
        if self.highlights_file and self.highlights_file.exists():
            highlights = cast(
                HighlightsType, json.loads(self.highlights_file.read_text())
            )
            return highlights
        else:
            return {}

    def add_highlight(
        self, name: str, data: HighlightType, overwrite: bool = True
    ) -> None:
        """
        Add a highlight to `self.highlights_file`.

        This is usually done behind the scenes using the `%%hlt My Title` magic.

        Parameters
        ----------
        name : str
            Name of highlight (also taken as the title).
        data : list
            list of data and metadata that can be passed to `IPython.display.publish_display_data` to render.
        overwrite : bool
            If `False` will raise an exception if a highlight of the same `name` exists. Default is `True`
        """
        highlights = self.get_highlights()

        if self.highlights_file and highlights is not None:
            highlights = highlights.copy()
        else:
            return

        if not overwrite and name in highlights:
            raise KeyError("Attempting to overwrite existing meta value")

        highlights[name] = data
        self.highlights_file.write_text(json.dumps(highlights), encoding="utf-8")

    def remove_highlight(self, name: str) -> None:
        """
        Remove highlight from highlight file. Performed by calling `get_highlights()`, then deleting the key `name` from
        the dictionary, then re-writing the highlights... if you're interested!
        """
        highlights = self.get_highlights()

        if not highlights or not self.highlights_file:
            return

        del highlights[name]
        self.highlights_file.write_text(json.dumps(highlights), encoding="utf-8")

    @cached_prop
    def href(self) -> Union[str, None]:
        """
        href usable in notebook HTML giving link to `self.file`.

        Assumes that `os.getcwd()` reflects the directory of the currently opened `.ipynb` (usually true, unless you're
        changing working dir).

        Does do escaping on the HTML, but is maybe pointless!

        Returns
        -------
        href : str
            href usable in notebook HTML.
        """
        return html.escape(Path(os.path.relpath(self.file, os.getcwd())).as_posix())

    def remove_files(self) -> None:
        """
        Deletes files associated with a `Tier`
        """
        if self.file:
            self.file.unlink()

        if self.meta_file:
            self.meta_file.unlink()


class HomeTierBase(FolderTierBase):
    """
    Home `Tier`.

    This, or a subclass of this should generally be the first entry in your hierarchy, essentially represents the top
    level folder in your hierarchy.

    Creates the `Home.ipynb` notebook that allows easy navigation of your project.
    """

    @cached_prop
    def name(self) -> str:
        return self.pretty_type

    @classmethod
    def iter_siblings(cls, parent: TierABC) -> Iterator[TierABC]:
        raise NotImplementedError("Home tier cannot be iterated over.")

    @cached_prop
    def folder(self) -> Path:
        assert self.child_cls
        return self.project.project_folder / (self.child_cls.pretty_type + "s")

    @cached_prop
    def file(self) -> Path:
        return self.project.project_folder / f"{self.name}.ipynb"

    def exists(self) -> bool:
        return bool(self.folder and self.file.exists())

    def setup_files(self, template: Union[Path, None] = None, meta=None) -> None:
        assert self.child_cls

        with FileMaker() as maker:
            print(f"Creating {self.child_cls.pretty_type} folder")
            maker.mkdir(self.folder, exist_ok=True)
            print("Success")

        with FileMaker() as maker:
            print(f"Creating Tier File ({self.file})")
            # TODO: look at this, is this ok?
            maker.write_file(
                self.file, (config.DEFAULT_TEMPLATE_DIR / "Home.ipynb").read_text()
            )
            print("Success")

    def remove_files(self) -> None:
        self.file.unlink()


class Project:
    """
    Represents your project. Understands your naming convention, and your project hierarchy.

    Parameters
    ----------
    hierarchy : Sequence[Type[BaseTier]]
        Sequence of `TierBase` subclasses representing the hierarchy for this project. i.e. earlier entries are stored
        in higher level directories.
    project_folder : Union[str, Path]
        path to home directory. Note this also accepts a path to a file, but will take `project_folder.parent` in that
        case. This enables `__file__` to be used if you want `project_folder` to be based in the same dir.


    Notes
    -----
    This class is a singleton i.e. only 1 instance per interpreter can be created.

    Project provides the following hooks, to allow customization of setting up and launching projects. These are
    lists of functions, which are called in order at the specified time:

        `__before_setup_files__` and `__after_setup_files__` - These are called when `project.setup_files()` is
        called. These can be used, for example, to create additional directories or files. These callables
        should take the project instance as an argument.

        `__before_launch__` and `__after_launch__` - These are called during and after `project.launch()` is
        called. These take the project and LabApp instance as an argument.
    """

    def __new__(cls, *args: Any, **kwargs: Any) -> Project:
        if env.project:
            raise RuntimeError(
                "Attempted to create new Project instance, only 1 instance permitted per interpreter"
            )
        instance = object.__new__(cls)
        env.project = instance
        return instance

    def __init__(
        self, hierarchy: Sequence[Type[TierABC]], project_folder: Union[str, Path]
    ):
        self._rank_map: Dict[Type[TierABC], int] = {}
        self._hierarchy: Sequence[Type[TierABC]] = []

        self.__before_setup_files__: List[Callable[[Project], None]] = []
        self.__after_setup_files__: List[Callable[[Project], None]] = []

        self.__before_launch__: List[
            Callable[[Project, Union[LabApp, None]], None]
        ] = []
        self.__after_launch__: List[Callable[[Project, Union[LabApp, None]], None]] = []

        self.hierarchy: Sequence[Type[TierABC]] = hierarchy

        project_folder_path = Path(project_folder).resolve()
        self.project_folder: Path = (
            project_folder_path
            if project_folder_path.is_dir()
            else project_folder_path.parent
        )

        self.template_env: PathLibEnv = PathLibEnv(
            autoescape=jinja2.select_autoescape(["html", "xml"]),
            loader=jinja2.FileSystemLoader(self.template_folder),
        )

    @property
    def hierarchy(self) -> Sequence[Type[TierABC]]:
        return self._hierarchy

    @hierarchy.setter
    def hierarchy(self, hierarchy: Sequence[Type[TierABC]]):
        self._hierarchy = hierarchy

        for rank, tier_cls in enumerate(hierarchy):
            self._rank_map[tier_cls] = rank

    @property
    def rank_map(self):
        return self._rank_map

    @property
    def home(self) -> TierABC:
        """
        Get the home `Tier`.
        """
        return self.hierarchy[0](project=self)

    def env(self, name: str) -> TierABC:
        """
        Initialise the global environment to a particular `Tier` that is retrieved by parsing `name`.

        This will set the value of `env.o`.

        Warnings
        --------
        This should only really be called once (or only with 1 name). Otherwise this could create some unexpected
        behaviour.
        """
        obj = self.__getitem__(name)

        if env.o and name != env.o.name:
            warn(
                (
                    f"Overwriting the global Tier {env.o} for this interpreter. This may cause unexpected behaviour. "
                    f"If you wish to create Tier objects that aren't the current Tier I recommend initialising them "
                    f"directly e.g. obj = MyTier('id1', 'id2')"
                )
            )

        env.update(obj)
        return obj

    def get_tier(self, identifiers: Tuple[str, ...]) -> TierABC:
        """
        Get a tier for a given set of identifiers.
        """
        cls = self.hierarchy[len(identifiers)]
        return cls(*identifiers, project=self)

    def get_child_cls(self, tier_cls: Type[TierABC]) -> Union[None, Type[TierABC]]:
        """
        Get the child class of a given tier class. Returns None if there is no child class
        """
        rank = self.rank_map[tier_cls]
        if rank + 1 > (len(self.hierarchy) - 1):
            return None
        else:
            cls = self.hierarchy[rank + 1]  # I don't understand why annotation needed?
            return cls

    def get_parent_cls(self, tier_cls: Type[TierABC]) -> Union[None, Type[TierABC]]:
        """
        Get the parent class of a given tier class. Returns None if there is no parent class
        """
        rank = self.rank_map[tier_cls]
        if rank - 1 < 0:
            return None
        else:
            cls = self.hierarchy[rank - 1]
            return cls

    def __getitem__(self, name: str) -> TierABC:
        """
        Retrieve a tier object from the project by name.

        Parameters
        ----------
        name : str
            Parsable name to get the tier object by. To get your `Home` just provide `Home.name`.

        Returns
        -------
        tier : TierBase
            Tier retrieved from project.
        """
        if name == self.home.name:
            obj = self.home
        else:
            identifiers = self.parse_name(name)
            if not identifiers:
                raise ValueError(f"Name {name} not recognised as identifying any Tier")
            obj = self.get_tier(identifiers)
        return obj

    @soft_prop
    def template_folder(self) -> Path:
        """
        Overwritable property providing where templates will be stored for this project.
        """
        return self.project_folder / "templates"

    def setup_files(self) -> TierABC:
        """
        Setup files needed for this project.

        Will put everything you need in `project_folder` to get going.
        """
        for func in self.__before_setup_files__:
            func(self)

        home = self.home

        if home.exists():
            return home

        print("Setting up project.")

        with FileMaker() as maker:
            print("Creating templates folder")
            maker.mkdir(self.template_folder)
            print("Success")

            for tier_cls in self.hierarchy:
                if not issubclass(tier_cls, NotebookTierBase):
                    continue

                maker.mkdir(self.template_folder / tier_cls.pretty_type)
                print("Copying over default template")
                maker.copy_file(
                    config.BASE_TEMPLATE,
                    self.template_folder / tier_cls.default_template,
                )
                print("Done")

        print("Setting up Home Tier")
        home.setup_files()
        print("Success")

        for func in self.__after_setup_files__:
            func(self)

        return home

    def launch(
        self, app: Union[LabApp, None] = None, patch_pythonpath: bool = True
    ) -> LabApp:
        """
        Jump off point for a cassini project.

        Sets up required files for your project, monkeypatches `PYTHONPATH` to make your project available throughout
        and launches a jupyterlab server.

        This explicitly associates an instance of the Jupyter server with a particular project.

        Parameters
        ----------
        app : LabApp
            A ready made Jupyter Lab app (By defuault will just create a new one).
        patch_pythonpath : bool
            Add `self.project_folder` to the `PYTHONPATH`? (defaults to `True`)
        """
        for func in self.__before_launch__:
            func(self, app)

        self.setup_files()

        if patch_pythonpath:
            py_path = os.environ.get("PYTHONPATH", "")
            project_path = str(self.project_folder.resolve())
            os.environ["PYTHONPATH"] = (
                py_path + os.pathsep + project_path if py_path else project_path
            )

        if app is None:
            app = CassiniLabApp()

        app.launch_instance()

        for func in self.__after_launch__:
            func(self, app)

        return app

    def parse_name(self, name: str) -> Tuple[str, ...]:
        """
        Parses a string that corresponds to a `Tier` and returns a list of its identifiers.

        returns an empty tuple if not a valid name.

        Parameters
        ----------
        name : str
            name to parse

        Returns
        -------
        identifiers : tuple
            identifiers extracted from name, empty tuple if `None` found

        Notes
        -----
        This works in a slightly strange - but robust way!

        e.g.

            >>> name = 'WP2.3c'

        it will loop through each entry in `cls.hierarchy` (skipping home!), and then perform a search on `name` with
        that regex:

            >>> WorkPackage.name_part_regex
            WP(\\d+)
            >>> match = re.search(WorkPackage.name_part_regex, name)

        If there's no match, it will return `()`, if there is, it stores the `id` part:

            >>> wp_id = match.group(1)  # in python group 0 is the whole match
            >>> wp_id
            2

        Then it removes the whole match from the name:

            >>> name = name[match.end(0):]
            >>> name
            .3c

        Then it moves on to the next tier

            >>> Experiment.name_part_regex
            '\\.(\\d+)'
            >>> match = re.search(WorkPackage.name_part_regex, name)

        If there's a match it extracts the id, and substracts the whole string from name and moves on, continuing this
        loop until it's gone through the whole hierarchy.

        The whole name has to be a valid id, or it will return `()` e.g.

            >>> TierBase.parse_name('WP2.3')
            ('2', '3')
            >>> TierBase.parse_name('WP2.u3')
            ()
        """
        parts = self.hierarchy[1:]
        ids: List[str] = []
        for tier_cls in parts:
            pattern = tier_cls.name_part_regex
            match = re.search(pattern, name)
            if match and match.start(0) == 0:
                ids.append(match.group(1))
                name = name[match.end(0) :]
            else:
                break
        if name:  # if there's any residual text then it's not a valid name!
            return tuple()
        else:
            return tuple(ids)

    def __repr__(self) -> str:
        return f"<Project at: '{self.project_folder}' hierarchy: '{self.hierarchy}' ({env})>"
